{{
        Fast Hartley Transform (FHT) in-place radix-2 DIT algorithm.

        Based on FHT function from the FXT library of algorithms by Jörg Arndt,
        also presented in his book "Matters Computational".

        This is almost 1:1 translation or original C++ code into PASM. I am sure
        there is room for optimisation - I think better use of cycles between
        hub memory accesses would improve performance.

        One of interesting features of this algorithm is that it only requires
        sine and cosine in first quadrant, making sine and cosine table lookup
        simplier and faster, and simplifying multiplication because multiplier
        is always above zero.

        Multipliers are treated as fixed point values below 1 in 0.12 notation.
        This makes implementation less prone to overflows without scaling at each
        stage.  For a transform of size N = 2^P, allowed input range is
        ±(16777216/N-1).
            for 8192 points transform, input range is ±2047
            for 4096 points transform, input range is ±4095
            for 2048 points transform, input range is ±8191
            for 1024 points transform, input range is ±16383
            for 512 points transform,  input range is ±32767
            for 256 points transform,  input range is ±65535
            for 128 points transform,  input range is ±131071

        This implementation's maximum transform size is 8192 (8192 * 4 = 32768),
        limited by amount of HUB RAM.  That is the extreme case that would require
        special care - entire program must be in assembly, no COG reloads can occur
        while preparing data for transform or doing the transform,
        and inter-COG signalling methods other than in-memory mailboxes should
        be used. So, 4096 should be considered as the  practical limit of transform
        size.


        Timing tests shown that execution takes longest time when input is a noise
        signal. After multiple tests, the following performance figures have been
        obtained (with about 3% allowance added) :

                    ┌───────────────────┬────────────────────────────────┐
                    │  Transform size   │ Worst case execution time (uS) │
                    ├───────────────────┼────────────────────────────────┤
                    │          16       │                   235          │
                    │          32       │                   570          │
                    │          64       │                  1350          │
                    │         128       │                  3100          │
                    │         256       │                  7000          │
                    │         512       │                 15500          │
                    │        1024       │                 34200          │
                    │        2048       │                 74500          │
                    └───────────────────┴────────────────────────────────┘

}}


{{

        Usage instructions for a transform of size N = 2^P:

        1. Start FHT cog using the start() method
        2. Fill an array of N longs with input data
        3. Perform the transform using the transform() method, passing these arguments:
                pBuffer - address of input array
                pBits - log2(N) = P
                pCommand - a combination of DECIMATE, BUTTERLFIES, POWER and MAGNITUDE constants:
                    DECIMATE -      perform decimation stage. Normally you should always specify
                                    this, unless your input is already in bit-reversed order
                    BUTTERLFIES -   perform the actual transform.Normally you should always specify
                                    this
                    POWER       -   calculate power of each spectral component
                    MAGNITUDE   -   calculate magnitude of each spectral component
                pWait - if TRUE, execution will be blocked until transform is completed.
                        if FALSE, transform() returns immediately

        5. If you passed FALSE in pWait, you can check whether the transform is completed
            using isBusy() method.
        6. If you do need more transforms, call stop() method to free the cog.
        7. On transform completion, the input array is overwritten with transform results.
           If MAGNITUDE was specified, array elements from 1 to N/2-1 contain magnitude of
           corresponding frequency component.
           If POWER was specified, array elements from 1 to N/2-1 contain power of
           corresponding frequency component.
           Relation between power and magnitude is the following : magnitude = sqrt(power).
           If either of MAGNITUDE or POWER was specified, array elements from N/2 to N-1 have
           no usable meaning after transform.

           If neither of MAGNITUDE or POWER was specified, the array contains raw result
           of discrete Hartley transform, and can be used, for example to derive a discrete
           Fourier transform.
        8. A convenience method oneshot() is provided, which is essentially a sequence of
           start(), transform() and stop()

}}


CON

    DECIMATE    = $01
    BUTTERLFIES = $02
    POWER       = $04
    MAGNITUDE   = $08

    O_COMMAND   = 0
    O_BUFFER    = 1
    O_BITS      = 2
    O_CYCLES    = 3

VAR
    long    gCog
    long    gParams[4]


'' starts the FHT cog
PUB start
    stop
    return gCog := (cognew(@fht_entry, @gParams)+1)

'' stops the FHT cog
PUB stop
    if gCog
        cogStop(gCog~ - 1)

{{
    This is the transform routine
}}

PUB transform(pBuffer, pBits, pCommand, pWait)
    wait
    gParams[O_BITS] := pBits
    gParams[O_BUFFER] := pBuffer
    gParams[O_COMMAND] := pCommand
    if pWait
        wait

PUB oneShot(pBuffer, pBits, pCommand)
    start
    gParams[O_BITS] := pBits
    gParams[O_BUFFER] := pBuffer
    gParams[O_COMMAND] := pCommand
    wait
    stop


PUB isBusy
    return gParams[O_COMMAND]

PUB wait
    repeat while gParams[O_COMMAND] <> 0

PUB getCycles
    return gParams[O_CYCLES]

DAT

fht_entry
                        movi    CTRA, # %11111 << 3
:wait
                        mov     cCommand, PAR
                        add     cCommand, #4
                        rdlong  cBuffer, cCommand
                        add     cCommand, #4
                        rdlong  P, cCommand
                        rdlong  cCommand, PAR            wz
        if_z            jmp     #:wait
                        neg     cycles, CNT

                        mov     len, #1
                        shl     len, P

'' =======================================================================================================
''                                      Bit reversal stage
'' =======================================================================================================
                        test    cCommand, #DECIMATE         wz
        if_z            jmp     #:no_decimate

                        neg     c2, P                       '' prepare operand for REV command
                        add     c2, #32

                        mov     c1, len                     '' sequence length
                        mov     t3, #0                      '' sequence index
                        mov     cAddress1, cBuffer          '' address of first element
:br
                        mov     cAddress2, t3               '' reverse bits
                        rev     cAddress2, c2
                        cmp     t3, cAddress2    wc         '' do not swap same pair twice
            if_nc       jmp     #:no_br
                        shl     cAddress2, #2               '' make address of second element
                        add     cAddress2, cBuffer
                        rdlong  t1, cAddress1               '' do the swap
                        rdlong  t2, cAddress2
                        wrlong  t2, cAddress1
                        wrlong  t1, cAddress2

:no_br                  add     cAddress1, #4               '' advance first address
                        add     t3, #1                      '' advance index
                        djnz    c1, #:br                    '' loop

:no_decimate

'' =======================================================================================================
''                                      Transform loop
'' +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
'' In all loops, array indexes are incremented by same number per loop. Instead of recalculating
'' memory addresses at each iteration, initial addresses are set up before loop, and are incremented
'' by appropriate number at each loop iteration.
'' =======================================================================================================
                        test    cCommand, #BUTTERLFIES         wz
        if_z            jmp     #:no_butterflies

                        mov     ldm, #1                     '' stage number, loop from 1 to P
                        mov     m, #2                       '' butterfly step
                        mov     mh, #1                      '' butterfly distance
                        mov     phi0, PI_                   '' sine/cosine arg
:loop1
                        mov     m4, mh
                        shr     m4, #1

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''                      Zero frequency and Nyquist frequency loops
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
                        mov     r, #0                               '' loop counter
                        mov     cAddress1, cBuffer                  '' r index
                        mov     cAddress2, mh                       '' r + mh index
                        shl     cAddress2, #2
                        add     cAddress2, cBuffer
                        mov     cAddress3, m4                       '' r + m4 index
                        mov     cAddress4, cAddress3
                        add     cAddress4, mh                       '' r + m4 + mh index
                        shl     cAddress3, #2
                        add     cAddress3, cBuffer
                        shl     cAddress4, #2
                        add     cAddress4, cBuffer
                        mov     astep, m                            '' address step, m * 4 per loop
                        shl     astep, #2
:loop2
                        '' sumdiff(f[r], f[r + mh])
                        rdlong  t1, cAddress1
                        rdlong  t2, cAddress2
                        add     t1, t2
                        wrlong  t1, cAddress1

                        sub     t1, t2
                        sub     t1, t2
                        wrlong  t1, cAddress2


                        '' skip Nyquist loop at first stage
                        cmp     m4, #0                  wz
        if_z            jmp     #:no_nyquist


                        '' sumdiff(f[r + m4], f[r + r4 + mh])
                        rdlong  t1, cAddress3
                        rdlong  t2, cAddress4
                        add     t1, t2
                        wrlong  t1, cAddress3

                        sub     t1, t2
                        sub     t1, t2
                        wrlong  t1, cAddress4
:no_nyquist
                        add     cAddress1, astep                    '' advance array indexes
                        add     cAddress2, astep
                        add     cAddress3, astep
                        add     cAddress4, astep
                        add     r, m                                '' advance loop var
                        cmp     r, len                  wc, wz      '' check loop condition
        if_b            jmp     #:loop2

                        mov     phi, phi0                           '' initialize sine/cosine argument
                        mov     j, #1                               '' setup loop control vars
                        mov     k, mh
                        sub     k, #1
                                                                    '' check initial loop condition
                        cmp     j, k                    wc, wz      '' j < k
        if_ae           jmp     #:skip

:loop3
                        '' get next sine/cosine pair
                        mov     dsin, phi
                        shl     dsin, #1                '' convert angle to table offset
                        mov     dcos, cos_table         '' for cosine, index from end
                        sub     dcos, dsin
                        add     dsin, sin_table         '' for sine, index from beginning
                        rdword  dsin, dsin
                        shr     dsin, #4
                        rdword  dcos, dcos
                        shr     dcos, #4


                        mov     cAddress1, mh                       '' r + mh + j index
                        add     cAddress1, j
                        mov     cAddress2, mh                       '' r + mh + k index
                        add     cAddress2, k
                        mov     cAddress3, j                        '' r + j index
                        mov     cAddress4, k                        '' r + k index

                        shl     cAddress1, #2
                        add     cAddress1, cBuffer
                        shl     cAddress2, #2
                        add     cAddress2, cBuffer
                        shl     cAddress3, #2
                        add     cAddress3, cBuffer
                        shl     cAddress4, #2
                        add     cAddress4, cBuffer

                        mov     astep, m                            '' address step
                        shl     astep, #2

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''                      Butterflies loop
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
                        mov     r, #0                               '' loop variable
:loop4
                        rdlong  t1, cAddress1                       '' fj = f[r + mh + j]
                        mov     m_2, t1
                        mov     m_1, dcos
                        rdlong  t2, cAddress2                       '' fk = f[r + mh + k]

                        '' f[r + mh + j] = fj * cos + fk * sin
                        call    #mult
                        mov     t3, m_3

                        mov     m_2, t2
                        mov     m_1, dsin
                        call    #mult

                        add     t3, m_3
                        wrlong  t3, cAddress1  '' no further read from cAddress1, value is kept in t3

                        '' f[r + mh + k] = fj * sin - fk * cos
                        mov     m_2, t1
                        mov     m_1, dsin
                        call    #mult
                        mov     t4, m_3

                        mov     m_2, t2
                        mov     m_1, dcos
                        call    #mult

                        sub     t4, m_3
                        wrlong  t4, cAddress2  '' no further read from cAddress2, value is kept in t4

                        '' sumdiff(f[r + j], f[r + mh + j])
                        rdlong  t1, cAddress3
                        add     t1, t3
                        wrlong  t1, cAddress3

                        sub     t1, t3
                        sub     t1, t3
                        wrlong  t1, cAddress1


                        '' sumdiff(f[r + k], f[r + mh + k])
                        rdlong  t1, cAddress4
                        add     t1, t4
                        wrlong  t1, cAddress4

                        sub     t1, t4
                        sub     t1, t4
                        wrlong  t1, cAddress2



                        add     cAddress1, astep                    '' advance indexes
                        add     cAddress2, astep
                        add     cAddress3, astep
                        add     cAddress4, astep
                        add     r, m                                '' advance loop var
                        cmp     r, len                  wc, wz      '' loop
        if_b            jmp     #:loop4

                        add     phi, phi0                           '' advance angle
                        add     j, #1                               '' advance loop vars
                        sub     k, #1
                        cmp     j, k                    wz, wc      '' loop
        if_b            jmp     #:loop3

:skip
                        shl     m, #1                               '' change butterfly step and distance
                        shl     mh, #1
                        shr     phi0, #1                            '' halve angle step
                        add     ldm, #1                             '' next stage
                        cmp     ldm, P                  wz, wc      '' loop
        if_be           jmp     #:loop1

:no_butterflies

'' =======================================================================================================
''                      Power/Magnitude calculation
'' =======================================================================================================
                        test    cCommand, # POWER | MAGNITUDE   wz  '' magnitude implies power
        if_z            jmp     #:no_power
                        mov     cAddress1, cBuffer
                        add     cAddress1, #4
                        mov     cAddress2, len
                        sub     cAddress2, #1
                        shl     cAddress2, #2
                        add     cAddress2, cBuffer
                        mov     c1, len
                        shr     c1, #1
                        sub     c1, #1
                        mov     c2, P
                        sub     c2, #1

:power
                        rdlong  m_1, cAddress1
                        sar     m_1, P
                        call    #mult2
                        mov     t1, m_3
                        rdlong  m_1, cAddress2
                        sar     m_1, P
                        call    #mult2
                        add     t1, m_3
                        shl     t1, #1
                        test    cCommand, #MAGNITUDE   wz
        if_z            mov     m_3, t1
        if_nz           call    #sqrt
                        wrlong  m_3, cAddress1
                        add     cAddress1, #4
                        sub     cAddress2, #4
                        djnz    c1, #:power
:no_power

:done
                        add     cycles, CNT
                        mov     t1, PAR
                        add     t1, #12
                        wrlong  cycles, t1
                        mov     t1, #0                              '' signal caller that we are done
                        wrlong  t1, PAR


                        jmp     #:wait



'' ************************************************************
''                      Multiplication
'' ************************************************************


mult                    
                        mov     m_3, #0                          '' initialize product
                        shl     m_1, #16  + 4                       '' shift miltiplier's MSB to bit 31
:loop                   shl     m_1, #1          wc, wz          '' get bit
        if_c            add     m_3, m_2                          '' add multiplicand if bit was one
        if_nz           sar     m_2, #1          wz              '' shift multiplicand
        if_nz           jmp     #:loop                          '' repeat if there are more '1' bits
                        sar     m_3, #1                          '' scale product
mult_ret                ret

mult2
                        mov     m_3, #0
                        abs     m_1, m_1
                        mov     m_2, m_1
:loop                   shr     m_1, #1          wc, wz
            if_c        add     m_3, m_2
                        shl     m_2, #1
            if_nz       jmp     #:loop
mult2_ret               ret


sqrt          mov       m_3,   #0                           'Reset root
              mov       m_2, h40000000                    'Reset mask (constant in register)
:sqloop       or        m_3, m_2                         'Set trial bit
              cmpsub    t1, m_3 wc                     'Subtract root from input if fits
              sumnc     m_3, m_2                         'Cancel trial bit, set root bit if fit
              shr       m_3, #1                           'Shift root down
              shr       m_2, #2                           'Shift mask down
              tjnz      m_2, #:sqloop                     'Loop until mask empty
sqrt_ret      ret
h40000000     long      $40000000


'' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
''                      Constants
'' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

PI_                     long    $1000
sin_table               long    $E000
cos_table               long    $F000

'' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
''                      Variables
'' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
astep                   res     1
j                       res     1
k                       res     1
r                       res     1
m4                      res     1
m                       res     1
mh                      res     1
phi0                    res     1
phi                     res     1
ldm                     res     1
dsin                    res     1
dcos                    res     1
P                       res     1
len                     res     1
cBuffer                 res     1
t1                      res     1
t2                      res     1
t3                      res     1
t4                      res     1
cAddress1               res     1
cAddress2               res     1
cAddress3               res     1
cAddress4               res     1
c1                      res     1
c2                      res     1
m_1                     res     1
m_2                     res     1
m_3                     res     1
m_4                     res     1
cCommand                res     1
cycles                  res     1

                        FIT     496
